
//  (C) Copyright Edward Diener 2011,2012
//  Use, modification and distribution are subject to the Boost Software License,
//  Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt).

#if !defined(BOOST_TTI_NAMESPACE_GEN_HPP)
#define BOOST_TTI_NAMESPACE_GEN_HPP

/*

  The succeeding comments in this file are in doxygen format.

*/

/** \file
*/

/// Generates the name of the Boost TTI namespace
/**
    returns = the generated name of the Boost TTI namespace.
*/
#define BOOST_TTI_NAMESPACE boost::tti

#endif // BOOST_TTI_NAMESPACE_GEN_HPP
